                  (* +--------------------------+
                     |  Listes  double entre  |
                     +--------------------------+ *)


type dbliste = { elem : int vect; mutable dbut : int; mutable lg : int};;


let premier l =
  if l.lg = 0 then failwith "file vide"
  else l.elem.(l.dbut)
;;

let dernier l =
  let n = vect_length(l.elem) in
  if l.lg = 0 then failwith "file vide"
  else l.elem.((l.dbut + l.lg - 1) mod n)
;;

let insre_dbut l x =
  let n = vect_length(l.elem) in
  if l.lg >= n then failwith "file pleine"
  else begin
    l.dbut <- (l.dbut + n - 1) mod n;
    l.elem.(l.dbut) <- x;
    l.lg <- l.lg+1;
  end
;;

let insre_fin l x =
  let n = vect_length(l.elem) in
  if l.lg >= n then failwith "file pleine"
  else begin
    l.elem.((l.dbut + l.lg) mod n) <- x;
    l.lg <- l.lg+1
  end
;;

let extrait_dbut l =
  let n = vect_length(l.elem) in
  if l.lg = 0 then failwith "file vide"
  else begin
    let x = l.elem.(l.dbut) in
    l.dbut <- (l.dbut + 1) mod n;
    l.lg <- l.lg - 1;
    x
  end
;;

let extrait_fin l =
  let n = vect_length(l.elem) in
  if l.lg = 0 then failwith "file vide"
  else begin
    let x = l.elem.((l.dbut + l.lg - 1) mod n) in
    l.lg <- l.lg - 1;
    x
  end
;;

                         (* +-------------+
                            |  Monotonie  |
                            +-------------+ *)

let croissant l =
  let n = vect_length(l.elem) and i = ref 0 in
  while (!i < l.lg - 1) &
        (l.elem.((l.dbut + !i) mod n) < l.elem.((l.dbut + !i + 1) mod n))
  do i := !i + 1 done;
  !i = l.lg - 1
;;

let dcroissant l =
  let n = vect_length(l.elem) and i = ref 0 in
  while (!i < l.lg - 1) &
        (l.elem.((l.dbut + !i) mod n) > l.elem.((l.dbut + !i + 1) mod n))
  do i := !i + 1 done;
  !i = l.lg - 1
;;

      


              (* +-----------------------------------+
                 |  Ralisabilit d'une permutation  |
                 +-----------------------------------+ *)

let copie l = { elem  = copy_vect l.elem; dbut = l.dbut; lg = l.lg };;

let rec recherche la lb lc =

  (* termin ? *)
  if (la.lg = 0) & (lb.lg = 0) then true

  (* gauche(lb) -> lc ? *)
  else if (lb.lg > 0) & (premier(lb) = lc.lg) then begin
    insre_fin lc (extrait_dbut lb);
    recherche la lb lc
  end

  (* droite(lb) -> lc ? *)
  else if (lb.lg > 0) & (dernier(lb) = lc.lg) then begin
    insre_fin lc (extrait_fin lb);
    recherche la lb lc
  end

  (* la est vide ? *)
  else if la.lg = 0 then false

  else begin

    (* optimisation :
       si la tte de la va dans lc au coup suivant ou si lb contient
       au plus un wagon, alors on peut insre systmatiquement  gauche
       dans lb (pas besoin de backtrack dans ce cas).
    *)
    let gauche,droite =
      if (lb.lg <= 1) or (premier(la) = lc.lg) then true,false
      else
        premier(la) < premier(lb) or dcroissant(lb),
        premier(la) < dernier(lb) or croissant(lb)
    in

    match gauche,droite with
      | false,false -> false

      | false,true  ->
	  insre_fin lb (extrait_dbut la);
          recherche la lb lc

      | true,false  ->
	  insre_dbut lb (extrait_dbut la);
          recherche la lb lc

      | true,true   ->
	  let sa = copie la and sb = copie lb and sc = copie lc in
	  insre_dbut lb (extrait_dbut la);
	  insre_fin   sb (extrait_dbut sa);
	  (recherche la lb lc) or (recherche sa sb sc)
  end
;;


                           (* +---------+
                              |  Essai  |
                              +---------+ *)

let random_perm n =
  let l = { elem = make_vect n 0; dbut = 0; lg = n } in
  for i=1 to n-1 do l.elem.(i) <- i done;
  for c = 1 to 2*n do
    let i = random__int(n) in
    let j = random__int(n) in
    let x = l.elem.(i) in l.elem.(i) <- l.elem.(j); l.elem.(j) <- x
  done;
  l
;;

let triable(l) =
  let n  = l.lg    in
  let la = copie l in
  let lb = {elem = make_vect n 0; dbut = 0; lg = 0} in
  let lc = {elem = make_vect n 0; dbut = 0; lg = 0} in
  recherche la lb lc
;;

                   (* +------------------------+
                      |  Liste des manoeuvres  |
                      +------------------------+ *)

type manoeuvre =
  | IG (* tte(A)   -> gauche(B) *)
  | ID (* tte(A)   -> droite(B) *)
  | XG (* gauche(B) -> queue(C)  *)
  | XD (* droite(B) -> queue(C)  *)
and rsultat =
  | Possible of manoeuvre list
  | Impossible
;;

let ajoute_manoeuvre m res = match res with
  | Possible(l) -> Possible(m::l)
  | Impossible     -> Impossible
;;

let rec cherche la lb lc =

  (* termin ? *)
  if (la.lg = 0) & (lb.lg = 0) then Possible([])

  (* gauche(lb) -> lc ? *)
  else if (lb.lg > 0) & (premier(lb) = lc.lg) then begin
    insre_fin lc (extrait_dbut lb);
    ajoute_manoeuvre XG (cherche la lb lc)
  end

  (* droite(lb) -> lc ? *)
  else if (lb.lg > 0) & (dernier(lb) = lc.lg) then begin
    insre_fin lc (extrait_fin lb);
    ajoute_manoeuvre XD (cherche la lb lc)
  end

  (* la est vide ? *)
  else if la.lg = 0 then Impossible

  else begin

    let gauche,droite =
      if (lb.lg <= 1) or (premier(la) = lc.lg) then true,false
      else
        premier(la) < premier(lb) or dcroissant(lb),
        premier(la) < dernier(lb) or croissant(lb)
    in

    match gauche,droite with
      | false,false -> Impossible

      | false,true  ->
	  insre_fin lb (extrait_dbut la);
          ajoute_manoeuvre ID (cherche la lb lc)

      | true,false  ->
	  insre_dbut lb (extrait_dbut la);
          ajoute_manoeuvre IG (cherche la lb lc)

      | true,true   ->
	  let sa = copie la and sb = copie lb and sc = copie lc in
	  insre_dbut lb (extrait_dbut la);
	  insre_fin   sb (extrait_dbut sa);
	  match cherche la lb lc with
	    | Possible(l) -> Possible(IG::l)
	    | Impossible ->  ajoute_manoeuvre ID (cherche sa sb sc)
  end
;;

let tri(l) =
  let n  = l.lg    in
  let la = copie l in
  let lb = {elem = make_vect n 0; dbut = 0; lg = 0} in
  let lc = {elem = make_vect n 0; dbut = 0; lg = 0} in
  cherche la lb lc
;;

             (* +-------------------------------------+
                |  Recherche des trains non triables  |
                +-------------------------------------+ *)


(* insre a  toutes les positions possibles dans l
   et retourne la liste des listes obtenues  *)
let rec insre a l = match l with
  | [] -> [[a]]
  | b::suite -> (a::l) :: (map (fun x -> b::x) (insre a suite))
;;

(* numre les permutations de [0,n-1] *)
let rec permutations(n) =
  if n <= 0   then []
  else if n=1 then [[0]]
  else flat_map (insre (n-1)) (permutations (n-1))
;;

(* cherche les trains de n wagons non triables *)
let non_triables(n) =
  find_all
    (fun l -> not(triable {elem = vect_of_list l; dbut = 0; lg = n}))
    (permutations n)
;;

(*
  non_triables(5) retourne
  [[3; 1; 2; 4; 0]; [1; 3; 2; 4; 0]; [4; 1; 2; 3; 0]; [1; 4; 2; 3; 0]]
*)
